import 'package:google_mobile_ads/google_mobile_ads.dart';
import 'package:knt_bloc/knt_bloc.dart';

import '../repository/monetized_repo.dart';
import '../service/admob_service.dart';

/// Handles generating & loading & showing [InterstitialAd] unit.
class FrequentlyAdsBloc extends BaseBloc {
  final AdmobService _admobService;
  final MonetizedRepo _monetizedRepo;

  InterstitialAd? _currentInterstitialAd;
  late int _retryLoadInterstitialAd;
  late int _interstitialAdCounter;

  /// Using this to define where ad request come from
  late String _currentAdTag;

  FrequentlyAdsBloc(this._admobService, this._monetizedRepo) {
    watch<BaseEvent>(
      _handleEvent,
      transformer: sequential(),
    );
  }

  @override
  void reset() {
    _currentInterstitialAd = null;
    _retryLoadInterstitialAd = 0;
    _interstitialAdCounter = 0;
    _currentAdTag = '';
  }

  @override
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(ClosedFrequentlyAdState(_currentAdTag));
  }

  Future<void> _handleEvent(
    BaseEvent event,
    Emitter<BaseState> emit,
  ) async {
    switch (event.runtimeType) {
      case PrepareFrequentlyAdEvent:
        return _handlePrepareFrequentlyAdEvent(
            event as PrepareFrequentlyAdEvent, emit);
      case ShowFrequentlyAdEvent:
        return _handleShowFrequentlyAdEvent(
            event as ShowFrequentlyAdEvent, emit);
      case _LoadDoneFrequentlyAdEvent:
        return _handleLoadDoneFrequentlyAdEvent(
            event as _LoadDoneFrequentlyAdEvent, emit);
      case _ReloadFrequentlyAdEvent:
        return _handleReloadFrequentlyAdEvent(
            event as _ReloadFrequentlyAdEvent, emit);
      case _CloseFrequentlyAdEvent:
        return _handleCloseFrequentlyAdEvent(
            event as _CloseFrequentlyAdEvent, emit);
    }
  }

  Future<void> _handlePrepareFrequentlyAdEvent(
    PrepareFrequentlyAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadInterstitialAd();
  }

  Future<void> _handleShowFrequentlyAdEvent(
    ShowFrequentlyAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    _currentAdTag = event.adTag;
    final purchasedIap = await _monetizedRepo.fetchPurchasedIapStatus();
    if (purchasedIap) {
      // Will not show any admobs when user has purchased
      emit(ClosedFrequentlyAdState(_currentAdTag));
      return;
    }
    // Also check for frequent of admobs
    _interstitialAdCounter++;
    if (_interstitialAdCounter < _admobService.thresholdShowInterstitialAds) {
      // If counter is not reached to threshold, then admob will not be shown
      emit(ClosedFrequentlyAdState(_currentAdTag));
      return;
    }
    // Otherwise load admobs and make money :D
    _interstitialAdCounter = 0;
    _retryLoadInterstitialAd = 0;
    if (_currentInterstitialAd == null) {
      await _createAndLoadInterstitialAd(forceLoadAndShow: true);
      return;
    }
    await _handleDisplayFrequentlyAdEvent(emit);
  }

  Future<void> _handleLoadDoneFrequentlyAdEvent(
    _LoadDoneFrequentlyAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleDisplayFrequentlyAdEvent(emit);
  }

  Future<void> _handleReloadFrequentlyAdEvent(
    _ReloadFrequentlyAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadInterstitialAd(forceLoadAndShow: event.forceShowAd);
  }

  Future<void> _handleCloseFrequentlyAdEvent(
    _CloseFrequentlyAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(ClosedFrequentlyAdState(_currentAdTag));
  }

  Future<void> _handleDisplayFrequentlyAdEvent(Emitter<BaseState> emit) async {
    if (_currentInterstitialAd == null) {
      return;
    }
    _currentInterstitialAd!.fullScreenContentCallback =
        FullScreenContentCallback(
      onAdShowedFullScreenContent: (ad) =>
          _log('$ad onAdShowedFullScreenContent.'),
      onAdDismissedFullScreenContent: (ad) {
        _log('$ad onAdDismissedFullScreenContent.');
        ad.dispose();
        add(_ReloadFrequentlyAdEvent());
      },
      onAdFailedToShowFullScreenContent: (ad, error) {
        _log('$ad onAdFailedToShowFullScreenContent: $error');
        ad.dispose();
        add(_ReloadFrequentlyAdEvent());
      },
    );
    await _currentInterstitialAd!.show();
    _currentInterstitialAd = null;
    emit(ClosedFrequentlyAdState(_currentAdTag));
  }

  Future<void> _createAndLoadInterstitialAd(
      {bool forceLoadAndShow = false}) async {
    _currentInterstitialAd?.dispose();
    final onLoaded = () {
      _retryLoadInterstitialAd = 0;
      if (forceLoadAndShow) {
        add(_LoadDoneFrequentlyAdEvent());
      }
    };
    final onLoadFailed = () {
      if (_retryLoadInterstitialAd < _admobService.maxRetryReloadAds) {
        _retryLoadInterstitialAd++;
        add(_ReloadFrequentlyAdEvent(forceShowAd: forceLoadAndShow));
        return;
      }
      // In case run out of retry, we close this ads flow
      if (forceLoadAndShow) {
        add(_CloseFrequentlyAdEvent());
      }
    };
    await InterstitialAd.load(
      adUnitId: _admobService.interstitialAdUnitId,
      request: AdRequest(),
      adLoadCallback: InterstitialAdLoadCallback(
        onAdLoaded: (InterstitialAd ad) {
          _currentInterstitialAd = ad;
          _currentInterstitialAd!.setImmersiveMode(true);
          onLoaded();
        },
        onAdFailedToLoad: (LoadAdError error) {
          onLoadFailed();
        },
      ),
    );
  }

  void _log(String message) => _admobService.loggingBuilder?.call(message);
}

// region EVENT
class PrepareFrequentlyAdEvent extends BaseEvent {}

class ShowFrequentlyAdEvent extends BaseEvent {
  final String adTag;

  const ShowFrequentlyAdEvent(this.adTag) : super(tag: adTag);
}

class _LoadDoneFrequentlyAdEvent extends BaseEvent {}

class _ReloadFrequentlyAdEvent extends BaseEvent {
  final bool forceShowAd;

  const _ReloadFrequentlyAdEvent({this.forceShowAd = false});
}

class _CloseFrequentlyAdEvent extends BaseEvent {}
// endregion

// region STATE
class ClosedFrequentlyAdState extends BaseState {
  final String adTag;

  const ClosedFrequentlyAdState(this.adTag) : super(tag: adTag);
}
// endregion
