import 'package:google_mobile_ads/google_mobile_ads.dart';
import 'package:knt_bloc/knt_bloc.dart';

import '../repository/monetized_repo.dart';
import '../service/admob_service.dart';

/// Handles generating & loading [BannerAd]/[NativeAd].
///
/// These Ads will be used to create [AdWidget].
class StaticAdsBloc extends BaseBloc {
  final AdmobService _admobService;
  final MonetizedRepo _monetizedRepo;

  BannerAd? _currentBannerAd;
  late int _retryLoadBannerAd;

  NativeAd? _currentNativeAd;
  late int _retryLoadNativeAd;

  StaticAdsBloc(this._admobService, this._monetizedRepo) {
    watch<BaseEvent>(
      _handleEvent,
      transformer: sequential(),
    );
  }

  @override
  void reset() {
    _retryLoadBannerAd = 0;
    _retryLoadNativeAd = 0;
  }

  Future<void> _handleEvent(
    BaseEvent event,
    Emitter<BaseState> emit,
  ) async {
    switch (event.runtimeType) {
      case PrepareBannerAdEvent:
        return _handlePrepareBannerAdEvent(event as PrepareBannerAdEvent, emit);
      case RefreshBannerAdEvent:
        return _handleRefreshBannerAdEvent(event as RefreshBannerAdEvent, emit);
      case _LoadDoneBannerAdEvent:
        return _handleLoadDoneBannerAdEvent(
            event as _LoadDoneBannerAdEvent, emit);

      case PrepareNativeAdEvent:
        return _handlePrepareNativeAdEvent(event as PrepareNativeAdEvent, emit);
      case RefreshNativeAdEvent:
        return _handleRefreshNativeAdEvent(event as RefreshNativeAdEvent, emit);
      case _LoadDoneNativeAdEvent:
        return _handleLoadDoneNativeAdEvent(
            event as _LoadDoneNativeAdEvent, emit);
    }
  }

  Future<void> _handlePrepareBannerAdEvent(
    PrepareBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadBannerAd();
  }

  Future<void> _handleRefreshBannerAdEvent(
    RefreshBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshBannerAd(emit);
  }

  Future<void> _handleLoadDoneBannerAdEvent(
    _LoadDoneBannerAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshBannerAd(emit);
  }

  Future<void> _handleRefreshBannerAd(Emitter<BaseState> emit) async {
    final purchasedIap = await _monetizedRepo.fetchPurchasedIapStatus();
    if (purchasedIap) {
      reset();
    }
    emit(LoadedBannerAdState(
      bannerAd: _currentBannerAd,
    ));
  }

  Future<void> _handlePrepareNativeAdEvent(
    PrepareNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _createAndLoadNativeAd();
  }

  Future<void> _handleRefreshNativeAdEvent(
    RefreshNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshNativeAd(emit);
  }

  Future<void> _handleLoadDoneNativeAdEvent(
    _LoadDoneNativeAdEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _handleRefreshNativeAd(emit);
  }

  Future<void> _handleRefreshNativeAd(
    Emitter<BaseState> emit,
  ) async {
    final purchasedIap = await _monetizedRepo.fetchPurchasedIapStatus();
    if (purchasedIap) {
      reset();
    }
    emit(LoadedNativeAdState(
      nativeAd: _currentNativeAd,
    ));
  }

  Future<void> _createAndLoadBannerAd({bool forceLoadAndShow = true}) async {
    _currentBannerAd?.dispose();
    final onLoaded = () {
      _retryLoadBannerAd = 0;
      if (forceLoadAndShow) {
        add(_LoadDoneBannerAdEvent());
      }
    };
    final onLoadFailed = () async {
      if (_retryLoadBannerAd < _admobService.maxRetryReloadAds) {
        _retryLoadBannerAd++;
        await _createAndLoadBannerAd();
      }
    };
    _currentBannerAd = _admobService.createBannerAd(
      onLoaded: onLoaded,
      onLoadFailed: onLoadFailed,
    );
    try {
      await _currentBannerAd!.load();
    } on Exception {
      onLoadFailed();
    }
  }

  Future<void> _createAndLoadNativeAd({bool forceLoadAndShow = true}) async {
    _currentNativeAd?.dispose();
    final onLoaded = () {
      _retryLoadNativeAd = 0;
      if (forceLoadAndShow) {
        add(_LoadDoneNativeAdEvent());
      }
    };
    final onLoadFailed = () async {
      if (_retryLoadNativeAd < _admobService.maxRetryReloadAds) {
        _retryLoadNativeAd++;
        await _createAndLoadNativeAd();
      }
    };
    _currentNativeAd = _admobService.createNativeAd(
      onLoaded: onLoaded,
      onLoadFailed: onLoadFailed,
    );
    try {
      await _currentNativeAd!.load();
    } on Exception {
      onLoadFailed();
    }
  }
}

// region EVENT
class PrepareBannerAdEvent extends BaseEvent {}

class RefreshBannerAdEvent extends BaseEvent {}

class _LoadDoneBannerAdEvent extends BaseEvent {}

class PrepareNativeAdEvent extends BaseEvent {}

class RefreshNativeAdEvent extends BaseEvent {}

class _LoadDoneNativeAdEvent extends BaseEvent {}
// endregion

// region STATE
class LoadedBannerAdState extends BaseState {
  final BannerAd? bannerAd;

  const LoadedBannerAdState({this.bannerAd});
}

class LoadedNativeAdState extends BaseState {
  final NativeAd? nativeAd;

  const LoadedNativeAdState({this.nativeAd});
}
// endregion
