import 'dart:async';

import 'package:flutter_inapp_purchase/flutter_inapp_purchase.dart';
import 'package:knt_bloc/knt_bloc.dart';

import '../model/subscription_item.dart';
import '../repository/monetized_repo.dart';
import '../service/iap_service.dart';

/// Handles in-app-purchase flow.
class SubscriptionBloc extends BaseBloc {
  final IapService _iapService;
  final MonetizedRepo _monetizedRepo;

  // In-app-purchase service
  late bool _purchased;
  late int _retryConnection;
  StreamSubscription? _purchaseUpdatedSubscription;
  StreamSubscription? _purchaseErrorSubscription;

  SubscriptionBloc(this._iapService, this._monetizedRepo) {
    watch<ConnectSubscriptionServiceEvent>(
      _handleConnectSubscriptionServiceEvent,
      transformer: droppable(),
    );
    watch<RequestSubscriptionEvent>(
      _handleRequestSubscriptionEvent,
      transformer: sequential(),
    );
    watch<RestoreSubscriptionEvent>(
      _handleRestoreSubscriptionEvent,
      transformer: restartable(),
    );
    watch<FetchListSkuEvent>(
      _handleFetchListSkuEvent,
      transformer: restartable(),
    );
    watch<_UpdatePurchaseInfoEvent>(
      _handleUpdatePurchaseInfoEvent,
      transformer: sequential(),
    );
    watch<_UpdateRestoreInfoEvent>(
      _handleUpdateRestoreInfoEvent,
      transformer: sequential(),
    );
    add(ConnectSubscriptionServiceEvent());
  }

  @override
  void reset() {
    _purchased = false;
    _retryConnection = 0;
  }

  @override
  Future<void> close() async {
    await FlutterInappPurchase.instance.finalize();
    _purchaseUpdatedSubscription?.cancel();
    _purchaseErrorSubscription?.cancel();
    return super.close();
  }

  @override
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    if (event.tag == (FetchListSkuEvent).toString()) {
      emit(FetchedListSkuFailureState());
      return;
    }
    if (event.tag == (RestoreSubscriptionEvent).toString()) {
      emit(FetchedRestoreFailureState());
      return;
    }
    if (event.tag == (RequestSubscriptionEvent).toString()) {
      emit(UnPurchasedState());
      return;
    }
    return super.handleBaseErrorEvent(event, emit);
  }

  Future<void> _handleConnectSubscriptionServiceEvent(
    ConnectSubscriptionServiceEvent event,
    Emitter<BaseState> emit,
  ) async {
    try {
      await FlutterInappPurchase.instance.initialize();
      // Setup callback for purchased actions
      _purchaseUpdatedSubscription?.cancel();
      _purchaseUpdatedSubscription =
          FlutterInappPurchase.purchaseUpdated.listen(
        (productItem) async {
          for (String sku in _skuList) {
            if (productItem?.productId == sku) {
              await _verifyPurchase(productItem);
              _purchased = true;
              add(_UpdatePurchaseInfoEvent());
              return;
            }
          }
        },
      );
      _purchaseErrorSubscription?.cancel();
      _purchaseErrorSubscription = FlutterInappPurchase.purchaseError.listen(
        (purchaseError) {
          _purchased = false;
          add(_UpdatePurchaseInfoEvent());
        },
      );
      // Get available subscriptions
      _fetchAvailableSubscription();
    } on Exception {
      if (_retryConnection < _iapService.maxRetryConnectBilling) {
        _retryConnection++;
        add(ConnectSubscriptionServiceEvent());
      } else {
        _purchased = false;
        add(_UpdatePurchaseInfoEvent());
      }
    }
  }

  Future<void> _handleRequestSubscriptionEvent(
    RequestSubscriptionEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(PurchasingState());
    List<IAPItem> itemList =
        await FlutterInappPurchase.instance.getSubscriptions(_offeringSkuList);
    for (var item in itemList) {
      if (item.productId == event.subItem.sku) {
        FlutterInappPurchase.instance.requestPurchase(item.productId ?? '');
        return;
      }
    }
  }

  Future<void> _handleRestoreSubscriptionEvent(
    RestoreSubscriptionEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(RestoringState());
    _fetchAvailableSubscription(forRestorePurpose: true);
  }

  Future<void> _handleFetchListSkuEvent(
    FetchListSkuEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(FetchingListSkuState());
    final itemList = await cancelableFuture(
      FlutterInappPurchase.instance.getSubscriptions(_offeringSkuList),
      tag: (FetchListSkuEvent).toString(),
    );
    final subList = _iapService.skuConverter(itemList);
    if (subList.isEmpty) {
      throw KntException(KntIssue.undefined);
    }
    emit(FetchedListSkuState(subList));
  }

  Future<void> _handleUpdatePurchaseInfoEvent(
    _UpdatePurchaseInfoEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _monetizedRepo.savePurchasedIapStatus(purchased: _purchased);
    if (_purchased) {
      emit(PurchasedState());
      return;
    }
    emit(UnPurchasedState());
  }

  Future<void> _handleUpdateRestoreInfoEvent(
    _UpdateRestoreInfoEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _monetizedRepo.savePurchasedIapStatus(purchased: _purchased);
    if (_purchased) {
      emit(FetchedRestoreSuccessState());
      return;
    }
    emit(FetchedRestoreFailureState());
  }

  Future<void> _fetchAvailableSubscription(
      {bool forRestorePurpose = false}) async {
    final purchasedItems = await cancelableFuture(
      FlutterInappPurchase.instance.getAvailablePurchases(),
      tag: (RestoreSubscriptionEvent).toString(),
    );
    _purchased = false;
    for (PurchasedItem item in purchasedItems ?? []) {
      if (item.productId == null) {
        continue;
      }
      if (_skuList.contains(item.productId)) {
        await _verifyPurchase(item);
        _purchased = true;
        break;
      }
    }
    if (forRestorePurpose) {
      add(_UpdateRestoreInfoEvent());
    } else {
      add(_UpdatePurchaseInfoEvent());
    }
  }

  Future<void> _verifyPurchase(PurchasedItem? item) async {
    if (item == null) {
      return;
    }
    await FlutterInappPurchase.instance.finishTransaction(item);
  }

  List<String> get _skuList => _iapService.skuList;

  List<String> get _offeringSkuList => _iapService.offeringSkuList;
}

// region EVENT
class ConnectSubscriptionServiceEvent extends BaseEvent {}

class RequestSubscriptionEvent extends BaseEvent {
  final SubscriptionItem subItem;

  const RequestSubscriptionEvent(this.subItem);

  @override
  String toString() => 'RequestSubscriptionEvent: item has sku ${subItem.sku}';
}

class RestoreSubscriptionEvent extends BaseEvent {}

class _UpdatePurchaseInfoEvent extends BaseEvent {}

class _UpdateRestoreInfoEvent extends BaseEvent {}

class FetchListSkuEvent extends BaseEvent {}
// endregion

// region STATE
class PurchasingState extends BaseState {}

class RestoringState extends BaseState {}

class PurchasedState extends BaseState {}

class UnPurchasedState extends BaseState {}

class FetchingListSkuState extends BaseState {}

class FetchedListSkuState extends BaseHasResultState<List<SubscriptionItem>> {
  const FetchedListSkuState(List<SubscriptionItem> subList) : super(subList);
}

class FetchedListSkuFailureState extends BaseErrorState {
  FetchedListSkuFailureState()
      : super(exception: KntException(KntIssue.undefined));
}

class FetchedRestoreSuccessState extends BaseState {}

class FetchedRestoreFailureState extends BaseState {}
// endregion
