import 'package:app_tracking_transparency/app_tracking_transparency.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:google_mobile_ads/google_mobile_ads.dart';

import '../model/admob_config.dart';

/// Service handles admob initialization & generates ad unit
class AdmobService {
  final AdmobConfig _admobConfig;

  const AdmobService(this._admobConfig);

  /// Initialize [MobileAds], must be called before [runApp].
  ///
  /// For example:
  /// ```
  /// Future<void> main() async {
  ///   WidgetsFlutterBinding.ensureInitialized();
  ///   await AdmobService().init();
  ///   runApp(App());
  /// }
  /// ```
  Future<void> init() async {
    // Need to request tracking app
    try {
      final trackingStatus =
          await AppTrackingTransparency.trackingAuthorizationStatus;
      if (trackingStatus == TrackingStatus.notDetermined) {
        await AppTrackingTransparency.requestTrackingAuthorization();
      }
    } on PlatformException {
      // Ignored
    }

    // If success, safely initialize admob
    await MobileAds.instance.initialize();
    await MobileAds.instance.setAppVolume(appVolume);
    await MobileAds.instance.updateRequestConfiguration(
        RequestConfiguration(testDeviceIds: testDeviceIds));
  }

  /// Generate [BannerAd] with callback functions, then use it to create [AdWidget] widget.
  ///
  /// For example:
  ///
  /// ```
  /// myBanner.load();
  /// final AdWidget adWidget = AdWidget(ad: myBanner);
  /// final Container adContainer = Container(
  ///   alignment: Alignment.center,
  ///   child: adWidget,
  ///   width: myBanner.size.width.toDouble(),
  ///   height: myBanner.size.height.toDouble(),
  /// );
  /// ```
  BannerAd createBannerAd({
    Function()? onLoaded,
    Function()? onLoadFailed,
  }) =>
      BannerAd(
        adUnitId: bannerAdUnitId,
        request: const AdRequest(),
        size: bannerSize,
        listener: BannerAdListener(
          onAdLoaded: (ad) => onLoaded?.call(),
          onAdFailedToLoad: (ad, error) {
            onLoadFailed?.call();
            loggingBuilder
                ?.call('Banner Ad load failed error - ${error.toString()}');
          },
        ),
      );

  /// Generate [NativeAd] with callback functions, then use this controller to create [NativeAd] widget.
  ///
  /// For example:
  ///
  /// ```
  /// myNative.load();
  /// final AdWidget adWidget = AdWidget(ad: myNative);
  /// final Container adContainer = Container(
  ///   alignment: Alignment.center,
  ///   child: adWidget,
  ///   width: 500,
  ///   height: 500,
  /// );
  /// ```
  NativeAd createNativeAd({
    Function()? onLoaded,
    Function()? onLoadFailed,
  }) =>
      NativeAd(
        adUnitId: nativeAdUnitId,
        factoryId: nativeAdFactoryId,
        request: const AdRequest(),
        listener: NativeAdListener(
          onAdLoaded: (ad) => onLoaded?.call(),
          onAdFailedToLoad: (ad, error) {
            onLoadFailed?.call();
            loggingBuilder
                ?.call('Native Ad load failed error - ${error.toString()}');
          },
        ),
      );

  String get nativeAdUnitId => _admobConfig.nativeAdUnitId ?? '';

  String get bannerAdUnitId => _admobConfig.bannerAdUnitId ?? '';

  String get interstitialAdUnitId => _admobConfig.interstitialAdUnitId ?? '';

  double get appVolume => _admobConfig.appVolume ?? 0.0;

  List<String> get testDeviceIds => _admobConfig.testDeviceIds;

  Function(String)? get loggingBuilder => _admobConfig.loggingBuilder;

  int get thresholdShowInterstitialAds =>
      _admobConfig.thresholdShowInterstitialAds;

  int get maxRetryReloadAds => _admobConfig.maxRetryReloadAds;

  AdSize get bannerSize => _admobConfig.bannerAdSize;

  String get nativeAdFactoryId => _admobConfig.nativeAdFactoryId;
}
